/*
  xselection.hpp - X selection command line interface

  Copyright (C) 2012 Taiji Yamada <taiji@aihara.co.jp>

  This is distributed under the MIT ("X") License.
 */
#ifndef _XSELECTION_HPP_
#define _XSELECTION_HPP_
#include <iostream>
#include <string>
#include <map>
#include <cstdio>
#include <cstdlib>
#include <unistd.h>	// fork, etc
#include <X11/Xatom.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Atoms.h>
#include <X11/Xmu/StdSel.h>
#include "buffer.hpp"
#include "xt_view.hpp"

template <typename B>
static
void storeBuffer(Widget w, XtPointer client_data,
                 Atom *selection, Atom *type, XtPointer value,
                 unsigned long *length, int *format)
{
  B *that = (B *)client_data;
  Display *d = XtDisplay(w);
  Atom *target = &that->target;
  if ((*type == 0 || *type == XT_CONVERT_FAIL) &&
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
      *target == XA_UTF8_STRING(d)
#else
      *target == XA_TEXT(d)
#endif
      ) {
    XtFree((char *)value);
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
    *target = XA_TEXT(d);
#else
    *target = XA_STRING;
#endif
    that->fetch_request();
    return;
  }
  {
    VERBOSE(std::cerr << *length          << " bytes fetched at " <<
            XmuGetAtomName(d, *selection) << " selection for " <<
            XmuGetAtomName(d, *target)    << " target, " <<
            XmuGetAtomName(d, *type)      << " type" << std::endl);
  }
  XTextProperty text_prop = { (unsigned char *)value, *type, *format, *length };
  int count, c;
  char **text_list = NULL;
  deject_if(XmbTextPropertyToTextList(d, &text_prop, &text_list, &count) != Success);
  XtFree((char *)value);
  for (c=0; c<count; ++c)
    that->data.append(text_list[c], strlen(text_list[c]));
  XFreeStringList(text_list);
  that->exit_flag = true;
}

template <typename B>
class xselection_base : public B {
public:
  buffer data;
  String selection_name;
  Atom selection;
  xselection_base(int &argc, char *argv[],
                  String selection_name_ = (String)"PRIMARY") :
    B(argc, argv), selection_name(selection_name_) {}
  xselection_base(int &argc, char *argv[], int w, int h,
                  String selection_name_ = (String)"PRIMARY") :
    B(argc, argv, w, h), selection_name(selection_name_) {}
};
template <typename B>
class xselection_getter : public xselection_base<B> {
public:
  Atom target;
  xselection_getter(int &argc, char *argv[],
                    String selection_name = (String)"PRIMARY") :
    xselection_base<B>(argc, argv, selection_name),
    target(
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
           XA_UTF8_STRING(XtDisplay(xselection_base<B>::toplevel))
#else
           XA_TEXT(XtDisplay(xselection_base<B>::toplevel))
#endif
           ) {}
  xselection_getter(int &argc, char *argv[], int w, int h,
                    String selection_name = (String)"PRIMARY") :
    xselection_base<B>(argc, argv, w, h, selection_name),
    target(
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
           XA_UTF8_STRING(XtDisplay(xselection_base<B>::toplevel))
#else
           XA_TEXT(XtDisplay(xselection_base<B>::toplevel))
#endif
           ) {}
  void fetch_request()
  {
    XmuInternStrings(XtDisplay(xselection_base<B>::toplevel),
                     &this->selection_name, (Cardinal)1, &this->selection);
    XtGetSelectionValue(xselection_base<B>::toplevel, xselection_base<B>::selection, target,
                        storeBuffer<xselection_getter<B> >, (XtPointer)this,
                        XtLastTimestampProcessed(XtDisplay(xselection_base<B>::toplevel)));
  }
  int loop()
  {
    fetch_request();
    xt_view::loop();
    store();
    return 0;
  }
  void store() const
  {
    std::cout.write(xselection_base<B>::data.ptr, xselection_base<B>::data.size);
    VERBOSE(std::cerr << xselection_base<B>::data.size << " bytes wrote to stdout" << std::endl);
  }
};
template <>
static
void storeBuffer<xselection_getter<xt_text_view>
		 >(Widget w, XtPointer client_data,
		   Atom *selection, Atom *type, XtPointer value,
		   unsigned long *length, int *format)
{
  xselection_getter<xt_text_view> *that = (xselection_getter<xt_text_view> *)client_data;
  Display *d = XtDisplay(w);
  Atom *target = &that->target;
  if ((*type == 0 || *type == XT_CONVERT_FAIL) &&
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
      *target == XA_UTF8_STRING(d)
#else
      *target == XA_TEXT(d)
#endif
      ) {
    XtFree((char *)value);
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
    *target = XA_TEXT(d);
#else
    *target = XA_STRING;
#endif
    that->fetch_request();
    return;
  }
  {
    VERBOSE(std::cerr << *length          << " bytes fetched at " <<
            XmuGetAtomName(d, *selection) << " selection for " <<
            XmuGetAtomName(d, *target)    << " target, " <<
            XmuGetAtomName(d, *type)      << " type" << std::endl);
  }
  XTextProperty text_prop = { (unsigned char *)value, *type, *format, *length };
  int count, c;
  char **text_list = NULL;
  deject_if(XmbTextPropertyToTextList(d, &text_prop, &text_list, &count) != Success);
  XtFree((char *)value);
  for (c=0; c<count; ++c)
    that->data.append(text_list[c], strlen(text_list[c]));
  XFreeStringList(text_list);
  that->text_list = text_list_alloc(that->data.ptr, that->data.size, &that->text_list_count);
  {
    const int h = that->baselineskip();
    const point<int> s(that->string_origin());
    int w = 0;
    {
      for (size_t i=0; i!=that->text_list_count; ++i) {
	size_t l = strlen(that->text_list[i]);
	w = std::max(w, that->tabified_string_width(that->text_list[i], l));
      }
      w += s.x*2;
    }
    that->pd = dimension<int>(w, h*(int)that->text_list_count);
  }
  that->draw();
}
template <>
inline
int xselection_getter<xt_text_view>::loop()
{
  fetch_request();
  xt_text_view::loop();
  store();
  return 0;
}

std::map<std::string, buffer> temp;

template <typename B>
class xselection_setter : public xselection_getter<B> {
public:
  xselection_setter(int &argc, char *argv[],
                    String selection_name = (String)"PRIMARY") :
    xselection_getter<B>(argc, argv, selection_name) {}
  xselection_setter(int &argc, char *argv[], int w, int h,
                    String selection_name = (String)"PRIMARY") :
    xselection_getter<B>(argc, argv, w, h, selection_name) {}
  int loop()
  {
    fetch();
    store_request();
    SILENT(if (fork()) exit(0));
    xt_view::loop();
    return 0;
  }
  void fetch()
  {
    char buf[1024];
    while (!feof(stdin)) {
      std::streamsize size = sizeof(buf);
      std::cin.read(buf, size);
      size = std::cin.gcount();
      deject_if(!temp[xselection_getter<B>::selection_name].append(buf, size));
    }
    VERBOSE(std::cerr << temp[xselection_getter<B>::selection_name].size << " bytes read from stdin" << std::endl);
    temp[xselection_getter<B>::selection_name].append("", 1);	// for ^@ char tails at prior x.org
  }
  void store_request()
  {
    Boolean convertSelection(Widget w, Atom *selection, Atom *target,
                             Atom *type, XtPointer *value,
                             unsigned long *length, int *format);
    void LoseSelection(Widget w, Atom *selection);
    XmuInternStrings(XtDisplay(xselection_getter<B>::toplevel), &this->selection_name,
                     (Cardinal)1, &this->selection);
    XtOwnSelection(xselection_getter<B>::toplevel, xselection_getter<B>::selection,
                   XtLastTimestampProcessed(XtDisplay(xselection_getter<B>::toplevel)),
                   convertSelection, LoseSelection, NULL);
  }
};
template <>
int xselection_setter<xt_text_view>::loop()
{
  fetch();
  store_request();
  text_list = text_list_alloc(temp[selection_name].ptr, temp[selection_name].size, &text_list_count);
  xt_text_view::loop();
  return 0;
}
Boolean convertSelection(Widget w, Atom *selection, Atom *target,
                         Atom *type, XtPointer *value,
                         unsigned long *length, int *format)
{
  Display *d = XtDisplay(w);
  XSelectionRequestEvent *req = XtGetSelectionRequest(w, *selection, NULL);
  String selection_name = XmuGetAtomName(d, *selection);
  if (
#if defined(USE_I18N) && defined(XA_COMPOUND_TEXT)
      *target == XA_COMPOUND_TEXT(d) ||
#endif
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
      *target == XA_UTF8_STRING(d) ||
#endif
      *target == XA_STRING || *target == XA_TEXT(d)) {
#if defined(USE_I18N) && defined(XA_COMPOUND_TEXT)
    if (
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
        *target == XA_UTF8_STRING(d) ||
        *target == XA_COMPOUND_TEXT(d) ||
#endif
        *target == XA_TEXT(d)) {
      size_t count;
      char **text_list = NULL;
      static const char *encoding_styles[] = {
        "XStringStyle", "XCompoundTextStyle", "XTextStyle", "XStdICCTextStyle",
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
        "XUTF8StringStyle",
#endif
      };
      XICCEncodingStyle style = XStringStyle;
      XTextProperty text_prop = { NULL, XA_NULL(d), 8, 0 };
      buffer data;
      deject_if(!(text_list = text_list_alloc(temp[selection_name].ptr, temp[selection_name].size, &count)));
      VERBOSE(std::cerr << count << " lines will be translated" << std::endl);
      // for XmbTextListToTextProperty has 1024-length limit at !*.UTF-8 locale
      if (*target == XA_TEXT(d)) {
        style = XCompoundTextStyle;
        *type = XA_COMPOUND_TEXT(d);
      }
#if defined(USE_UTF8) && X_HAVE_UTF8_STRING
      else if (*target == XA_UTF8_STRING(d)) {
        style = XUTF8StringStyle;
        *type = XA_UTF8_STRING(d);
      }
      else if (*target == XA_COMPOUND_TEXT(d)) {
        style = XUTF8StringStyle;
        *type = XA_UTF8_STRING(d);
      }
#endif
      deject_if(XmbTextListToTextProperty(d, text_list, count,
                                          style, &text_prop) != Success);
      text_list_free(text_list, count);
      data.text_append((char *)text_prop.value, text_prop.nitems);
      // for ^@ char mingles with text for '\0'-separated buffer at Emacs
      XFree(text_prop.value);
      *length = data.size;
      deject_if(!(*value = XtMalloc(*length)));
      memcpy((char *)*value, data.ptr, *length);
      *format = text_prop.format;
      VERBOSE(std::cerr << "text property of " << encoding_styles[style] << ": " <<
              "encoding: " << XmuGetAtomName(d, text_prop.encoding) << ", " <<
              "format: "   << text_prop.format << ", " <<
              "nitems: "   << text_prop.nitems << std::endl);
    }
    else /*if (*target == XA_STRING)*/
#endif
    {
      *type = XA_STRING;
      *length = temp[selection_name].size - 1;	// for ^@ char tails at prior x.org
      deject_if(!(*value = XtMalloc(*length)));
      memcpy((char *)*value, temp[selection_name].ptr, *length);
      *format = 8;
    }
    VERBOSE(std::cerr << *length          << " bytes stored at " <<
            XmuGetAtomName(d, *selection) << " selection for " <<
            XmuGetAtomName(d, *target)    << " target, " <<
            XmuGetAtomName(d, *type)      << " type" << std::endl);
    return True;
  }
  else
    VERBOSE(std::cerr << "unexpected target: " << XmuGetAtomName(d, *target) << std::endl);
  if (XmuConvertStandardSelection(w, req->time, selection, target,
                                  type, (XPointer *)value, length, format))
    return True;
  return False;
}
void LoseSelection(Widget w, Atom *selection)
{
  Display *d = XtDisplay(w);
  String selection_name = XmuGetAtomName(d, *selection);
  temp.erase(std::string(selection_name));
  if (temp.size()) return;
  XtCloseDisplay(XtDisplay(w));
  exit(0);
}
template <typename B>
class xselection_adder : public xselection_setter<B> {
public:
  xselection_adder(int &argc, char *argv[],
                   String selection_name = (String)"PRIMARY") :
    xselection_setter<B>(argc, argv, selection_name) {}
  xselection_adder(int &argc, char *argv[], int w, int h,
                   String selection_name = (String)"PRIMARY") :
    xselection_setter<B>(argc, argv, w, h, selection_name) {}
  int loop()
  {
    xselection_setter<B>::fetch_request();
    B::loop();
    deject_if(!temp[xselection_setter<B>::selection_name].append(xselection_setter<B>::data.ptr, xselection_setter<B>::data.size));
    xselection_setter<B>::data.clear();
    xselection_setter<B>::exit_flag = false;
    SILENT(if (fork()) exit(0));
    xselection_setter<B>::loop();
    return 0;
  }
};
template <typename B>
class xselection_continuing_setter : public xselection_setter<B> {
public:
  xselection_continuing_setter(int &argc, char *argv[],
                               String selection_name = (String)"PRIMARY") :
    xselection_setter<B>(argc, argv, selection_name) {}
  int loop()
  {
    xselection_setter<B>::store_request();
    char buf[64];	// small buffer for binary std::cin of pipe
    std::ios::sync_with_stdio(false);
    while (!xselection_setter<B>::exit_flag) {
      while (XtAppPending(xselection_setter<B>::app_context)) {
        XtAppNextEvent(xselection_setter<B>::app_context, &this->event);
        XtDispatchEvent(&this->event);
      }
      size_t size = temp[xselection_setter<B>::selection_name].size;
      struct timeval tv = { 0, 1000000/10 };
      struct timeval tv_start, tv_stop, tv_ZERO = { 0, 0 };
      gettimeofday(&tv_start, NULL);
      while (!std::cin.eof()) {
        std::streamsize size = sizeof(buf);
        std::cin.read(buf, size);
        size = std::cin.gcount();
        if (size)
          deject_if(!temp[xselection_setter<B>::selection_name].append(buf, size));
        gettimeofday(&tv_stop, NULL);
        if (tv < tv_stop - tv_start) { tv = tv_ZERO; break; }
      }
      if (temp[xselection_setter<B>::selection_name].size > size) {
        VERBOSE(std::cerr << temp[xselection_setter<B>::selection_name].size << " bytes read from stdin" << std::endl);
      }
      if (select(0, NULL, NULL, NULL, &tv) == -1) {}
      std::cin.clear();
    }
    return 0;
  }
};
template <typename B>
class xselection_following_setter : public xselection_continuing_setter<B> {
public:
  xselection_following_setter(int &argc, char *argv[],
                              String selection_name = (String)"PRIMARY") :
    xselection_continuing_setter<B>(argc, argv, selection_name) {}
  int loop()
  {
    std::cin.seekg(0, std::ios::end);
    std::cin.clear();
    return xselection_continuing_setter<B>::loop();
  }
};
template <typename B>
class xselection_continuing_adder : public xselection_continuing_setter<B> {
public:
  xselection_continuing_adder(int &argc, char *argv[],
                              String selection_name = (String)"PRIMARY") :
    xselection_continuing_setter<B>(argc, argv, selection_name) {}
  int loop()
  {
    xselection_continuing_setter<B>::fetch_request();
    B::loop();
    deject_if(!temp[xselection_continuing_setter<B>::selection_name].append(xselection_continuing_setter<B>::data.ptr, xselection_continuing_setter<B>::data.size));
    xselection_continuing_setter<B>::data.clear();
    xselection_continuing_setter<B>::exit_flag = false;
    SILENT(if (fork()) exit(0));
    xselection_continuing_setter<B>::loop();
    return 0;
  }
};
template <typename B>
class xselection_following_adder : public xselection_following_setter<B> {
public:
  xselection_following_adder(int &argc, char *argv[],
                             String selection_name = (String)"PRIMARY") :
    xselection_following_setter<B>(argc, argv, selection_name) {}
  int loop()
  {
    xselection_following_adder<B>::fetch_request();
    B::loop();
    deject_if(!temp[xselection_following_setter<B>::selection_name].append(xselection_following_setter<B>::data.ptr, xselection_following_setter<B>::data.size));
    xselection_following_setter<B>::data.clear();
    xselection_following_setter<B>::exit_flag = false;
    SILENT(if (fork()) exit(0));
    xselection_following_setter<B>::loop();
    return 0;
  }
};
template <typename B>
class xselection_rotator : public xselection_setter<B> {
public:
  int &argc;
  char **argv;
  int rotation_number;
  xselection_rotator(int &argc_, char *argv_[],
                     int rotation_number_ = 1) :
    xselection_setter<B>(argc_, argv_, NULL),
    argc(argc_), argv(argv_), rotation_number(rotation_number_) {}
  int loop()
  {
    xselection_getter<B> v[] = {
      xselection_getter<B>(argc, argv, (String)"PRIMARY"),
      xselection_getter<B>(argc, argv, (String)"SECONDARY"),
      xselection_getter<B>(argc, argv, (String)"CLIPBOARD"),
    };
    size_t n = sizeof(v)/sizeof(v[0]);
    size_t i, j;
    for (i=0; i<n; ++i) {
      v[i].fetch_request();
      v[i].B::loop();
    }
    for (i=0; i<n; ++i) {
      j = (rotation_number + ((rotation_number < 0) ? (-rotation_number/n+1)*n : 0) + i) % n;
      xselection_setter<B>::selection_name = v[j].selection_name;
      temp[xselection_setter<B>::selection_name].append(v[i].data.ptr, v[i].data.size);
      VERBOSE(std::cerr << temp[xselection_setter<B>::selection_name].size << " bytes of " <<
              v[i].selection_name << " selection will be stored at " <<
              v[j].selection_name << " selection" << std::endl);
      xselection_setter<B>::store_request();
    }
    SILENT(if (fork()) exit(0));
    B::loop();
    return 0;
  }
};
template <typename B>
class xselection_swapper : public xselection_setter<B> {
public:
  int &argc;
  char **argv;
  xselection_swapper(int &argc_, char *argv_[]) :
    xselection_setter<B>(argc_, argv_, NULL),
    argc(argc_), argv(argv_) {}
  int loop()
  {
    xselection_getter<B> v[] = {
      xselection_getter<B>(argc, argv, (String)"PRIMARY"),
      xselection_getter<B>(argc, argv, (String)"SECONDARY"),
    };
    const size_t n = sizeof(v)/sizeof(v[0]);
    size_t i, j;
    for (i=0; i<n; ++i) {
      v[i].fetch_request();
      v[i].B::loop();
    }
    const int rotation_number = 1;
    for (i=0; i<n; ++i) {
      j = (rotation_number + ((rotation_number < 0) ? (-rotation_number/n+1)*n : 0) + i) % n;
      xselection_setter<B>::selection_name = v[j].selection_name;
      temp[xselection_setter<B>::selection_name].append(v[i].data.ptr, v[i].data.size);
      VERBOSE(std::cerr << temp[xselection_setter<B>::selection_name].size << " bytes of " <<
              v[i].selection_name << " selection will be stored at " <<
              v[j].selection_name << " selection" << std::endl);
      xselection_setter<B>::store_request();
    }
    SILENT(if (fork()) exit(0));
    B::loop();
    return 0;
  }
};
#endif
