/*
  x11_view.hpp

  Copyright (C) 2012 Taiji Yamada <taiji@aihara.co.jp>

  This is distributed under the MIT ("X") License.
*/
#ifndef _X11_VIEW_HPP_
#define _X11_VIEW_HPP_
#include <iostream>
#include <string>
#include <cstdio>
#include <cstdlib>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#if defined(USE_XFT)
#include <freetype/config/ftheader.h>
#include <X11/Xft/Xft.h>
#endif
#if defined(USE_CAIRO)
#include <cairo.h>
#include <cairo-xlib.h>
#endif
#include <sys/select.h>	// select, etc
#include <sys/time.h>	// gettimeofday, etc
#include "verbose.h"
#include "deject.h"
#include "text_list.h"
#include "timeval.hpp"
#include "geometry.hpp"

class x11_view_base {
public:
  char *locale;
  Display *display;
  x11_view_base(int &argc, char *argv[]) : locale(NULL)
  {
    (void)argc, (void)argv;
    deject_if(!(display = XOpenDisplay(NULL)));
#if defined(USE_I18N) || (defined(USE_UTF8) && X_HAVE_UTF8_STRING)
    if ((locale = setlocale(LC_ALL, "")))
      if (XSupportsLocale())
        XSetLocaleModifiers("");
#endif
  }
  virtual ~x11_view_base()
  {
    XCloseDisplay(display);
  }
};

class x11_view : public x11_view_base {
public:
  x11_view(int &argc, char *argv[],
           int w = 0, int h = 0, bool double_buffer = false, long event_mask = 0,
	   const char *font_name = NULL) :
    x11_view_base(argc, argv), pixmap((XID)0)
  {
    screen_number = DefaultScreen(display);
    parent_window = RootWindow(display, screen_number);
    XSetWindowAttributes attributes;
    unsigned long valuemask = 0;
    attributes.background_pixel = WhitePixel(display, screen_number);
    valuemask |= CWBackPixel;
    attributes.border_pixel = BlackPixel(display, screen_number);
    valuemask |= CWBorderPixel;
    attributes.bit_gravity = NorthWestGravity;
    valuemask |= CWBitGravity;
    if (event_mask) {
      attributes.event_mask = event_mask;
      valuemask |= CWEventMask;
    }
    window = XCreateWindow(display, parent_window,
			   0, 0, w ? w : 1, h ? h : 1, 0,
			   DefaultDepth(display, screen_number),
			   InputOutput,
			   DefaultVisual(display, screen_number),
			   valuemask, &attributes);
    gc = DefaultGC(display, screen_number);
#if /*defined(USE_XFT) ||*/ defined(USE_CAIRO)
    double_buffer = true;
#endif
    if (double_buffer) {
      pixmap = XCreatePixmap(display, window, w ? w : 1, h ? h : 1,
			     DefaultDepth(display, screen_number));
      XSetForeground(display, gc, WhitePixel(display, screen_number));
      XFillRectangle(display, pixmap, gc, 0, 0, w ? w : 1, h ? h : 1);
    }
    XSetForeground(display, gc, BlackPixel(display, screen_number));
#if defined(USE_CAIRO)
    cr_surface = cairo_xlib_surface_create(display, (pixmap != (XID)0) ? pixmap : window,
					   DefaultVisual(display, screen_number),
					   w ? w : 1, h ? h : 1);
    cr = cairo_create(cr_surface);
    cairo_select_font_face( cr, font_name ? font_name :
			    (std::string(locale).find_first_of("ja") != std::string::npos) ?
			    "IPAGothic"/*"MSGothic"*//*"Meiryo"*//*"Hiragino Gothic Pro"*//*"Kozuka Gothic Pr6N"*/ :
			    /*"Monaco"*//*"Menlo"*/"Arial Unicode MS"
			    , CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
    cairo_set_font_size(cr, 13);
    cairo_font_extents(cr, &cr_font_extents);
    cairo_set_line_width(cr, 1);
#endif
#if defined(USE_XFT)
    {
      ft_font = XftFontOpen(display, screen_number,
			    XFT_FAMILY, XftTypeString, font_name ? font_name :
			    (std::string(locale).find_first_of("ja") != std::string::npos) ?
			    "IPAGothic"/*"MSGothic"*//*"Meiryo"*//*"Hiragino Gothic Pro"*//*"Kozuka Gothic Pr6N"*/ :
			    /*"Monaco"*//*"Menlo"*/"Arial Unicode MS"
			    , XFT_SIZE, XftTypeInteger, 10, NULL);
      ft_draw = XftDrawCreate(display, (pixmap != (XID)0) ? pixmap : window,
			      DefaultVisual(display, screen_number),
			      DefaultColormap(display, screen_number));
      XftColorAllocName(display,
			DefaultVisual(display, screen_number),
			DefaultColormap(display, screen_number), "Black", &ft_fgcolor);
      XftColorAllocName(display,
			DefaultVisual(display, screen_number),
			DefaultColormap(display, screen_number), "White", &ft_bgcolor);
    }
#endif
#if defined(USE_I18N) || (defined(USE_UTF8) && X_HAVE_UTF8_STRING)
    {
      char **missing_charset_list;
      int missing_charset_count;
      char *def_string_return; 
      font_set = XCreateFontSet(display, font_name ? font_name :
				"-*-*-medium-r-normal--10-*-*-*-*-*-*-*",
				&missing_charset_list, &missing_charset_count,
				&def_string_return);
      font_set_extents = XExtentsOfFontSet(font_set);
    }
#endif
    if (!font_name)
      font_struct = XQueryFont(display, XGContextFromGC(gc));
    else {
      font_struct = XLoadQueryFont(display, font_name);
      XSetFont(display, gc, font_struct->fid);
    }
    if (w && h) XMapWindow(display, window);
    XFlush(display);
  }
  ~x11_view()
  {
    XUnmapWindow(display, window);
#if defined(USE_CAIRO)
    cairo_destroy(cr);
    cairo_surface_destroy(cr_surface);
#endif
#if defined(USE_XFT)
    XftFontClose(display, ft_font);
    XftDrawDestroy(ft_draw);
    XftColorFree(display,
		 DefaultVisual(display, screen_number),
		 DefaultColormap(display, screen_number), &ft_fgcolor);
    XftColorFree(display,
		 DefaultVisual(display, screen_number),
		 DefaultColormap(display, screen_number), &ft_bgcolor);
#endif
#if defined(USE_I18N) || (defined(USE_UTF8) && X_HAVE_UTF8_STRING)
    XFreeFontSet(display, font_set);
#endif
  }
  const rectangle<int, unsigned int> get_geometry() const
  {
    Window root;
    rectangle<int, unsigned int> g;
    unsigned int border_width, depth; 
    XGetGeometry(display, (pixmap != (XID)0) ? pixmap : window,
		 &root, &g.p.x, &g.p.y, &g.d.w, &g.d.h,
		 &border_width, &depth);
    return g;
  }
  void flush() const
  {
    if (pixmap != (XID)0) {
      rectangle<int, unsigned int> g = get_geometry();
      XCopyArea(display, pixmap, window, gc,
		0, 0, g.d.w, g.d.h, 0, 0);
    }
    XFlush(display);
  }
  void rewind() const
  {
    if (pixmap != (XID)0) {
      rectangle<int, unsigned int> g = get_geometry();
      XGCValues values;
      XGetGCValues(display, gc, GCForeground | GCBackground, &values);
      XSetForeground(display, gc, values.background);
      XFillRectangle(display, pixmap, gc, 0, 0, g.d.w ? g.d.w : 1, g.d.h ? g.d.h : 1);
      XSetForeground(display, gc, values.foreground);
      return;
    }
    XClearWindow(display, window);
  }
  void resize_procedure(dimension<unsigned int> d)
  {
    if (pixmap != (XID)0) {
      XFreePixmap(display, pixmap);
      pixmap = XCreatePixmap(display, window, d.w ? d.w : 1, d.h ? d.h : 1,
			     DefaultDepth(display, screen_number));
      XGCValues values;
      XGetGCValues(display, gc, GCForeground | GCBackground, &values);
      XSetForeground(display, gc, values.background);
      XFillRectangle(display, pixmap, gc, 0, 0, d.w ? d.w : 1, d.h ? d.h : 1);
      XSetForeground(display, gc, values.foreground);
#if defined(USE_CAIRO)
      cairo_xlib_surface_set_drawable(cr_surface, pixmap, d.w ? d.w : 1, d.h ? d.h : 1);
#endif
#if defined(USE_XFT)
      XftDrawChange(ft_draw, pixmap);
#endif
    }
#if defined(USE_CAIRO)
    cairo_xlib_surface_set_size(cr_surface, d.w, d.h);
#endif
  }
  void swap_ground_colors()
  {
#if defined(USE_XFT)
    std::swap(ft_fgcolor, ft_bgcolor);
#endif
    XGCValues values;
    XGetGCValues(display, gc, GCForeground | GCBackground, &values);
    std::swap(values.foreground, values.background);
    XChangeGC(display, gc, GCForeground | GCBackground, &values);
#if defined(USE_CAIRO)
    XColor fgcolor;
    fgcolor.pixel = values.foreground;
    XQueryColor(display, DefaultColormap(display, screen_number), &fgcolor);
    cairo_set_source_rgb(cr,
			 (fgcolor.flags & DoRed)   ? fgcolor.red   : 0,
			 (fgcolor.flags & DoGreen) ? fgcolor.green : 0,
			 (fgcolor.flags & DoBlue)  ? fgcolor.blue  : 0);
#endif   
  }
  const int baselineskip() const {
#if defined(USE_CAIRO)
    return cr_font_extents.height;
#elif defined(USE_XFT)
    return ft_font->height;
#elif defined(USE_I18N) || (defined(USE_UTF8) && X_HAVE_UTF8_STRING)
    return font_set_extents->max_logical_extent.height;
#else
    return font_struct->max_bounds.ascent + font_struct->max_bounds.descent;
#endif
  }
  const point<int> string_origin() const
  {
#if defined(USE_CAIRO)
    return point<int>(1, cr_font_extents.ascent);
#elif defined(USE_XFT)
    return point<int>(1, ft_font->height - ft_font->descent - 1);
#elif defined(USE_I18N) || (defined(USE_UTF8) && X_HAVE_UTF8_STRING)
    return point<int>(-font_set_extents->max_logical_extent.x + 1, font_set_extents->max_logical_extent.height-1 - 1);
#else
    return point<int>(font_struct->max_bounds.lbearing, font_struct->max_bounds.ascent);
#endif
  }
  void draw_string(const point<int> &p, const char *string, const int num_bytes) const
  {
    const point<int> s(string_origin());
    const int n = num_bytes - ((string[num_bytes-1] == '\n') ? 1 : 0);
#if defined(USE_CAIRO)
    cairo_move_to(cr, p.x+s.x, p.y+s.y);
    cairo_show_text(cr, std::string(string, n).c_str());
#elif defined(USE_XFT)
    XftDrawStringUtf8(ft_draw, &ft_fgcolor, ft_font, p.x+s.x, p.y+s.y, (FcChar8 *)string, n);
#elif defined(USE_I18N) || (defined(USE_UTF8) && X_HAVE_UTF8_STRING)
    Drawable d = (pixmap != (XID)0) ? pixmap : window;
#if defined(FORCE_UTF8)
    Xutf8DrawString(display, d, font_set, gc, p.x+s.x, p.y+s.y, string, n);
#else
    XmbDrawString(display, d, font_set, gc, p.x+s.x, p.y+s.y, string, n);
#endif
#else
    Drawable d = (pixmap != (XID)0) ? pixmap : window;
    XDrawString(display, d, gc, p.x+s.x, p.y+s.y, string, n);
#endif
  }
  const int string_width(const char *string, const int num_bytes) const
  {
    int w;
    const int n = num_bytes - ((string[num_bytes-1] == '\n') ? 1 : 0);
#if defined(USE_CAIRO)
    cairo_text_extents_t cr_text_extents;
    cairo_text_extents(cr, std::string(string, n).c_str(), &cr_text_extents);
    w = cr_text_extents.x_advance;
#elif defined(USE_XFT)
    XGlyphInfo ft_extents;
    XftTextExtentsUtf8(display, ft_font, (FcChar8 *)string, n, &ft_extents);
    w = ft_extents.xOff;
#elif defined(USE_I18N) || (defined(USE_UTF8) && X_HAVE_UTF8_STRING)
    XRectangle overall_ink;
    XRectangle overall_logical; 
#if defined(FORCE_UTF8)
    Xutf8TextExtents(font_set, string, n, &overall_ink, &overall_logical);
#else
    XmbTextExtents(font_set, string, n, &overall_ink, &overall_logical);
#endif
    w = overall_logical.width;
#else
    w = XTextWidth(font_struct, string, n);
#endif
    return w;
  }
  const rectangle<int> string_rectangle(const point<int> &p, const char *string, const int num_bytes) const
  {
    const point<int> s(string_origin());
    const int n = num_bytes - ((string[num_bytes-1] == '\n') ? 1 : 0);
    rectangle<int> r;
#if defined(USE_CAIRO)
    cairo_text_extents_t cr_text_extents;
    cairo_text_extents(cr, std::string(string, n).c_str(), &cr_text_extents);
    r = rectangle<int>(p.x+s.x+cr_text_extents.x_bearing, p.y+s.y+cr_text_extents.y_bearing, cr_text_extents.width, cr_text_extents.height);
#elif defined(USE_XFT)
    XGlyphInfo extents;
    XftTextExtentsUtf8(display, ft_font, (FcChar8 *)string, n, &extents);
    r = rectangle<int>(p.x+s.x-extents.x, p.y+s.y-extents.y, extents.width, extents.height);
#elif defined(USE_I18N) || (defined(USE_UTF8) && X_HAVE_UTF8_STRING)
    XRectangle overall_ink;
    XRectangle overall_logical; 
#if defined(FORCE_UTF8)
    Xutf8TextExtents(font_set, string, n, &overall_ink, &overall_logical);
#else
    XmbTextExtents(font_set, string, n, &overall_ink, &overall_logical);
#endif
    r = rectangle<int>(p.x+s.x+overall_ink.x, p.y+s.y+overall_ink.y, overall_ink.width, overall_ink.height);
#else
    r = rectangle<int>(p.x, p.y, XTextWidth(font_struct, string, n)+s.x, baselineskip());
#endif
    return r;
  }
  void draw_tabified_string(point<int> p, const char *string, const int num_bytes) const
  {
    const int n = num_bytes - ((string[num_bytes-1] == '\n') ? 1 : 0);
    char tab_skip[9];
    int i = 0, j = 0, l = 0;
    memset(tab_skip, ' ', sizeof(tab_skip)-1);
    tab_skip[sizeof(tab_skip)-1] = '\0';
    do {
      if (string[i] == '\t' || i == n-1) {
	int c = i-j+((string[i] == '\t') ? 0 : 1);
	draw_string(p, &string[j], c);
	p.x += string_width(&string[j], c);
	l += c;
	if (string[i] == '\t') {
	  c = 8 - l%8;
	  draw_string(p, tab_skip, c);
	  p.x += string_width(tab_skip, c);
	  l += c;
	}
	j = i+1;
      }
    } while (++i < n);
  }
  const int tabified_string_width(const char *string, const int num_bytes) const
  {
    const int n = num_bytes - ((string[num_bytes-1] == '\n') ? 1 : 0);
    char tab_skip[9];
    int i = 0, j = 0, l = 0, w = 0;
    memset(tab_skip, ' ', sizeof(tab_skip)-1);
    tab_skip[sizeof(tab_skip)-1] = '\0';
    do {
      if (string[i] == '\t' || i == n-1) {
	int c = i-j+((string[i] == '\t') ? 0 : 1);
	w += string_width(&string[j], c);
	l += c;
	if (string[i] == '\t') {
	  c = 8 - l%8;
	  w += string_width(tab_skip, c);
	  l += c;
	}
	j = i+1;
      }
    } while (++i < n);
    return w;
  }
  const rectangle<int> tabified_string_rectangle(point<int> p, const char *string, const int num_bytes) const
  {
    const int n = num_bytes - ((string[num_bytes-1] == '\n') ? 1 : 0);
    std::pair<rectangle<int>, bool> r;
    char tab_skip[9];
    int i = 0, j = 0, l = 0;
    memset(tab_skip, ' ', sizeof(tab_skip)-1);
    tab_skip[sizeof(tab_skip)-1] = '\0';
    do {
      if (string[i] == '\t' || i == n-1) {
	int c = i-j+((string[i] == '\t') ? 0 : 1);
	if (r.second)
	  r.first += string_rectangle(p, &string[j], c);
	else {
	  r.first = string_rectangle(p, &string[j], c);
	  r.second = true;
	}
	p.x += string_width(&string[j], c);
	l += c;
	if (string[i] == '\t') {
	  c = 8 - l%8;
	  r.first += string_rectangle(p, tab_skip, c);
	  p.x += string_width(tab_skip, c);
	  l += c;
	}
	j = i+1;
      }
    } while (++i < n);
    return r.first;
  }
  void draw_rectangle(const rectangle<int> &r) const
  {
    Drawable d = (pixmap != (XID)0) ? pixmap : window;
    XDrawRectangle(display, d, gc, r.p.x, r.p.y, r.d.w, r.d.h);
  }
  void fill_rectangle(const rectangle<int> &r) const
  {
    Drawable d = (pixmap != (XID)0) ? pixmap : window;
    XFillRectangle(display, d, gc, r.p.x, r.p.y, r.d.w, r.d.h);
  }
protected:
  int screen_number;
  Window parent_window, window;
  Pixmap pixmap;
  GC gc;
  XFontStruct *font_struct;
#if defined(USE_I18N) || (defined(USE_UTF8) && X_HAVE_UTF8_STRING)
  XFontSet font_set;
  XFontSetExtents *font_set_extents;
#endif
#if defined(USE_XFT)
  XftFont *ft_font;
  XftDraw *ft_draw;
  XftColor ft_fgcolor, ft_bgcolor;
#endif
#if defined(USE_CAIRO)
  cairo_surface_t *cr_surface;
  cairo_t *cr;
  cairo_font_extents_t cr_font_extents;
#endif
};

#ifndef _wheelPredicate_
#define _wheelPredicate_
static
Bool wheelPredicate(Display *display, XEvent *event, XPointer arg)
{
  (void)display;
  XEvent *ev = (XEvent *)arg;
  return ((event->type == ButtonPress || event->type == ButtonRelease) &&
	  event->xbutton.button == ev->xbutton.button);
}
#endif

class x11_text_view : public x11_view {
public:
  size_t text_list_count;
  char **text_list;
  rectangle<int> cvp, ovp;
  dimension<int> pd;
  rectangle<int> sb[2];
  bool sb_flag[2];
  XEvent event;
  x11_text_view(int &argc, char *argv[], int w = 0, int h = 0) :
    x11_view(argc, argv, w, h, /* .double_buffer = */false,
	     KeyPressMask | KeyReleaseMask |
	     ButtonPressMask | ButtonReleaseMask | Button1MotionMask |
	     StructureNotifyMask,
	     NULL), text_list_count(0), text_list(NULL)
  {
    ovp = cvp = rectangle<int>(0, 0, w, h);
  }
  const int scale_sb_h(const rectangle<int> &vp, const int x) const
  {
    const int h = baselineskip(), w = h, m = 1;
    return x * (vp.d.w - w - m*2) / std::max(pd.w, vp.d.w);
  }
  const int rescale_sb_h(const rectangle<int> &vp, const int x) const
  {
    const int h = baselineskip(), w = h, m = 1;
    return x * std::max(pd.w, vp.d.w) / (vp.d.w - w - m*2);
  }
  const rectangle<int> sb_h(const rectangle<int> &vp) const
  {
    const int h = baselineskip(), w = h, m = 1;
    rectangle<int> r(scale_sb_h(vp, -vp.p.x) + m, vp.d.h - 1 - w + m,
		     scale_sb_h(vp, vp.d.w) - m*2, w - m*2);
    if (r.d.w < 3) { r.d.w = 3; --r.p.y; }
    return r;
  }
  const int scale_sb_v(const rectangle<int> &vp, const int y) const
  {
    const int h = baselineskip(), w = h, m = 1;
    return y * (vp.d.h - w - m*2) / std::max(pd.h, vp.d.h);
  }
  const int rescale_sb_v(const rectangle<int> &vp, const int y) const
  {
    const int h = baselineskip(), w = h, m = 1;
    return y * std::max(pd.h, vp.d.h) / (vp.d.h - w - m*2);
  }
  const rectangle<int> sb_v(const rectangle<int> &vp) const
  {
    const int h = baselineskip(), w = h, m = 1;
    rectangle<int> r(vp.d.w - 1 - w + m, scale_sb_v(vp, -vp.p.y) + m,
		     w - m*2, scale_sb_v(vp, vp.d.h) - m*2);
    if (r.d.h < 3) { r.d.h = 3; --r.p.x; }
    return r;
  }
  void draw()
  {
    if (!text_list_count) return;
    const int h = baselineskip();
    const rectangle<int> rects[][2] = {
      { sb_h(ovp), sb_v(ovp) },
      { sb_h(cvp), sb_v(cvp) },
    };
    int i, b = std::min(-ovp.p.y, -cvp.p.y)/h - 2,
      e = (cvp.d.h + std::max(-ovp.p.y, -cvp.p.y))/h + 2;
    if (pixmap != (XID)0) {
      rewind();
      {
	for (i=std::max(0, b); i<std::min((int)text_list_count, e); ++i) {
	  size_t l = strlen(text_list[i]);
	  draw_tabified_string(point<int>(cvp.p.x, cvp.p.y+h*i), text_list[i], l);
	}
      }
      for (i=0; i<2; ++i)
	draw_rectangle(rects[1][i]);
    }
    else {
      swap_ground_colors();
      for (i=0; i<2; ++i)
	draw_rectangle(rects[0][i]);
      swap_ground_colors();
      if (ovp.p.y < cvp.p.y)
	for (i=std::min((int)text_list_count, e)-1; i>=std::max(0, b); --i) {
	  size_t l = strlen(text_list[i]);
	  swap_ground_colors();
	  fill_rectangle(tabified_string_rectangle(point<int>(ovp.p.x, ovp.p.y+h*i), text_list[i], l));
	  swap_ground_colors();
	  draw_tabified_string(point<int>(cvp.p.x, cvp.p.y+h*i), text_list[i], l);
	}
      else
	for (i=std::max(0, b); i<std::min((int)text_list_count, e); ++i) {
	  size_t l = strlen(text_list[i]);
	  swap_ground_colors();
	  fill_rectangle(tabified_string_rectangle(point<int>(ovp.p.x, ovp.p.y+h*i), text_list[i], l));
	  swap_ground_colors();
	  draw_tabified_string(point<int>(cvp.p.x, cvp.p.y+h*i), text_list[i], l);
	}
      for (i=0; i<2; ++i)
	draw_rectangle(rects[1][i]);
    }
    flush();
    sb[0] = rects[1][0];
    sb[1] = rects[1][1];
    ovp = cvp;
  }
  void animate_inertia(const point<int> &speed, point<int> &dp, Time &dt, const int pending_count = 0)
  {
    point<int> decay[2] = { point<int>(9, 9), point<int>(10, 10) };
    if (pd.w < cvp.d.w || cvp.p.x > 0) dp.x = -cvp.p.x*speed.x;
    if (pd.h < cvp.d.h || cvp.p.y > 0) dp.y = -cvp.p.y*speed.y;
    if (!(pd.w < cvp.d.w) && cvp.p.x < -(pd.w - cvp.d.w)) dp.x = -(cvp.p.x + (pd.w - cvp.d.w))*speed.x;
    if (!(pd.h < cvp.d.h) && cvp.p.y < -(pd.h - cvp.d.h)) dp.y = -(cvp.p.y + (pd.h - cvp.d.h))*speed.y;
    while (dp != point<int>(0, 0)) {
      if (pd.w < cvp.d.w || cvp.p.x > 0) decay[0].x = 1, decay[1].x = 2;
      if (pd.h < cvp.d.h || cvp.p.y > 0) decay[0].y = 1, decay[1].y = 2;
      if (!(pd.w < cvp.d.w) && cvp.p.x < -(pd.w - cvp.d.w)) decay[0].x = 1, decay[1].x = 2;
      if (!(pd.h < cvp.d.h) && cvp.p.y < -(pd.h - cvp.d.h)) decay[0].y = 1, decay[1].y = 2;
      dp = dp*decay[0]/decay[1];
      cvp.p += dp/speed;
      draw();
      if (XPending(display) > pending_count) break;
      struct timeval tv = { 0, 1000/2*dt/std::max(speed.x, speed.y) };
      select(0, NULL, NULL, NULL, &tv);
    }
    if (pd.w < cvp.d.w || cvp.p.x > 0) cvp.p.x = 0;
    if (pd.h < cvp.d.h || cvp.p.y > 0) cvp.p.y = 0;
    if (!(pd.w < cvp.d.w) && cvp.p.x < -(pd.w - cvp.d.w)) cvp.p.x = -(pd.w - cvp.d.w);
    if (!(pd.h < cvp.d.h) && cvp.p.y < -(pd.h - cvp.d.h)) cvp.p.y = -(pd.h - cvp.d.h);
    draw();
  }
  int loop()
  {
    const int h = baselineskip();
    const point<int> s(string_origin());
    int w = 0;
    {
      for (size_t i=0; i!=text_list_count; ++i) {
	size_t l = strlen(text_list[i]);
	w = std::max(w, tabified_string_width(text_list[i], l));
      }
      w += s.x*2;
    }
    pd = dimension<int>(w, h*(int)text_list_count);
    w = string_width(" ", 1);

    point<int> op(0, 0), dp(0, 0);
    const point<int> key_speed(2, 2), motion_speed(2, 10), wheel_speed(10, 10);
    Time ot = 0, dt = 0;
    KeySym keysym;
    while (!0) {
      XNextEvent(display, &event);
      switch (event.type) {
      case KeyPress:
	ot = event.xkey.time;
	break;
      case KeyRelease:
	switch (keysym = XKeycodeToKeysym(display, event.xkey.keycode,
					  (event.xkey.state&ShiftMask)?1:0)) {
	case XK_Escape:	return 0; break;
	case XK_Left: case XK_Right: case XK_Up: case XK_Down:
	case XK_Home: case XK_End: case XK_Prior: case XK_Next:
	  dp = point<int>(0, 0);
	  switch (keysym) {
	  case XK_Left:  dp.x += w; break;
	  case XK_Up:    dp.y += h; break;
	  case XK_Right: dp.x -= w; break;
	  case XK_Down:  dp.y -= h; break;
	  case XK_Home:  dp.x += cvp.d.w; break;
	  case XK_Prior: dp.y += cvp.d.h; break;
	  case XK_End:   dp.x -= cvp.d.w; break;
	  case XK_Next:  dp.y -= cvp.d.h; break;
          }
	  cvp.p += dp;
	  dt = event.xkey.time - ot;
	  draw();
	  dp = point<int>(0, 0);
	  animate_inertia(key_speed, dp, dt);
	  break;
	}
	break;
      case ButtonPress:
	if (!(event.xbutton.button == Button1)) continue;
	op = point<int>(event.xbutton.x, event.xbutton.y);
	ot = event.xbutton.time;
	sb_flag[0] = sb[0].is_inner(op);
	sb_flag[1] = sb[1].is_inner(op);
	break;
      case ButtonRelease:
	if (!(event.xbutton.button == Button1)) {
	  int c = 1;
	  switch (event.xbutton.button) {
	  case Button1: case Button2: case Button3: continue;
	  default:
	    dp = point<int>((event.xbutton.button == 6) ? -1 : (event.xbutton.button == 7) ? +1 : 0,
			    (event.xbutton.button == 4) ? -1 : (event.xbutton.button == 5) ? +1 : 0);
	    cvp.p += dp;
	    draw();
	    ot = event.xbutton.time;
	    while (XCheckIfEvent(display, &event, wheelPredicate, (XPointer)&event)) ++c;
	    dt = std::max(1UL, event.xbutton.time - ot);
	    dp *= c;
	    if (XPending(display)) continue;
	    animate_inertia(wheel_speed/((pixmap != (XID)0) ? 5 : 1), dp, dt);
	    break;
	  }
	  continue;
	}
	if (sb_flag[0])
	  dp = point<int>(rescale_sb_h(cvp, -(event.xbutton.x - op.x)), 0);
	if (sb_flag[1])
	  dp = point<int>(0, rescale_sb_v(cvp, -(event.xbutton.y - op.y)));
	if (sb_flag[0] || sb_flag[1]) {
	  cvp.p += dp;
	  draw();
	  op = point<int>(event.xbutton.x, event.xbutton.y);
	  dt = event.xbutton.time - ot;
	  animate_inertia(wheel_speed, dp, dt);
	  continue;
	}
	dp = point<int>(event.xbutton.x - op.x, event.xbutton.y - op.y);
	cvp.p += dp;
	op = point<int>(event.xbutton.x, event.xbutton.y);
	dt = event.xbutton.time - ot;
	draw();
	dp *= motion_speed;
	animate_inertia(motion_speed/((pixmap != (XID)0) ? 2 : 1), dp, dt, (pixmap != (XID)0) ? 10 : 0);
	break;
      case MotionNotify:
	if (!(event.xmotion.state & Button1Mask)) continue;
	if (sb_flag[0])
	  dp = point<int>(rescale_sb_h(cvp, -(event.xbutton.x - op.x)), 0);
	if (sb_flag[1])
	  dp = point<int>(0, rescale_sb_v(cvp, -(event.xbutton.y - op.y)));
	if (sb_flag[0] || sb_flag[1]) {
	  cvp.p += dp;
	  draw();
	  op = point<int>(event.xbutton.x, event.xbutton.y);
	ot = event.xmotion.time;
	  continue;
	}
	dp = point<int>(event.xmotion.x - op.x, event.xmotion.y - op.y);
	cvp.p += dp;
	op = point<int>(event.xmotion.x, event.xmotion.y);
	ot = event.xmotion.time;
	draw();
	break;
      case ConfigureNotify:
	cvp.d = dimension<int>(event.xconfigure.width, event.xconfigure.height);
	resize_procedure(cvp.d);
	draw();
	break;
      }
    }
    return 0;
  }
};
#endif
