/*
  xcutbuffer.hpp - X cut buffer command line interface

  Copyright (C) 2012 Taiji Yamada <taiji@aihara.co.jp>

  This is distributed under the MIT ("X") License.
*/
#ifndef _XCUTBUFFER_HPP_
#define _XCUTBUFFER_HPP_
#include <iostream>
#include <string>
#include <cstdio>
#include <cstdlib>
#include "buffer.hpp"
#include "x11_view.hpp"

template <typename B>
class xcutbuffer_base : public B {
public:
  int buffer_index;
  buffer data;
  xcutbuffer_base(int &argc, char *argv[],
                  int buffer_index_ = 0) :
    B(argc, argv), buffer_index(buffer_index_) {}
  xcutbuffer_base(int &argc, char *argv[], int w, int h,
                  int buffer_index_ = 0) :
    B(argc, argv, w, h), buffer_index(buffer_index_) {}
};
template <typename B>
class xcutbuffer_getter : public xcutbuffer_base<B> {
public:
  xcutbuffer_getter(int &argc, char *argv[],
                    int buffer_index = 0) :
    xcutbuffer_base<B>(argc, argv, buffer_index) {}
  xcutbuffer_getter(int &argc, char *argv[], int w, int h,
                    int buffer_index = 0) :
    xcutbuffer_base<B>(argc, argv, w, h, buffer_index) {}
  int loop()
  {
    fetch();
    store();
    return 0;
  }
  void fetch()
  {
    int nbytes;
    xcutbuffer_base<B>::data.ptr = XFetchBuffer(xcutbuffer_base<B>::display, &nbytes, xcutbuffer_base<B>::buffer_index);
    xcutbuffer_base<B>::data.size = nbytes;
    VERBOSE(std::cerr << xcutbuffer_base<B>::data.size << " bytes fetched from " <<
            "cut buffer #" << xcutbuffer_base<B>::buffer_index << std::endl);
  }
  void store() const
  {
    std::cout.write(xcutbuffer_base<B>::data.ptr, xcutbuffer_base<B>::data.size);
    VERBOSE(std::cerr << xcutbuffer_base<B>::data.size << " bytes wrote to stdout" << std::endl);
  }
};
template <>
inline
int xcutbuffer_getter<x11_text_view>::loop()
{
  fetch();
  text_list = text_list_alloc(data.ptr, data.size, &text_list_count);
  x11_text_view::loop();
  store();
  return 0;
}
template <typename B>
class xcutbuffer_setter : public xcutbuffer_getter<B> {
public:
  xcutbuffer_setter(int &argc, char *argv[],
                    int buffer_index = 0) :
    xcutbuffer_getter<B>(argc, argv, buffer_index) {}
  xcutbuffer_setter(int &argc, char *argv[], int w, int h,
                    int buffer_index = 0) :
    xcutbuffer_getter<B>(argc, argv, w, h, buffer_index) {}
  int loop()
  {
    fetch();
    store();
    return 0;
  }
  void fetch()
  {
    char buf[1024];
    while (!feof(stdin)) {
      std::streamsize size = sizeof(buf);
      std::cin.read(buf, size);
      size = std::cin.gcount();
      deject_if(!xcutbuffer_getter<B>::data.append(buf, size));
    }
    VERBOSE(std::cerr << xcutbuffer_getter<B>::data.size << " bytes read from stdin" << std::endl);
  }
  void store() const
  {
    XStoreBuffer(xcutbuffer_getter<B>::display, xcutbuffer_getter<B>::data.ptr, xcutbuffer_getter<B>::data.size, xcutbuffer_getter<B>::buffer_index);
    VERBOSE(std::cerr << xcutbuffer_getter<B>::data.size << " bytes stored into " <<
            "cut buffer #" << xcutbuffer_getter<B>::buffer_index << std::endl);
  }
};
template <>
inline
int xcutbuffer_setter<x11_text_view>::loop()
{
  fetch();
  text_list = text_list_alloc(data.ptr, data.size, &text_list_count);
  x11_text_view::loop();
  store();
  return 0;
}
template <typename B>
class xcutbuffer_adder : public xcutbuffer_setter<B> {
public:
  xcutbuffer_adder(int &argc, char *argv[],
                   int buffer_index = 0) :
    xcutbuffer_setter<B>(argc, argv, buffer_index) {}
  xcutbuffer_adder(int &argc, char *argv[], int w, int h,
                   int buffer_index = 0) :
    xcutbuffer_setter<B>(argc, argv, w, h, buffer_index) {}
  int loop()
  {
    xcutbuffer_getter<B>::fetch();
    xcutbuffer_setter<B>::loop();
    return 0;
  }
};
template <typename B>
class xcutbuffer_continuing_setter : public xcutbuffer_setter<B> {
public:
  xcutbuffer_continuing_setter(int &argc, char *argv[],
                               int buffer_index = 0) :
    xcutbuffer_setter<B>(argc, argv, buffer_index) {}
  xcutbuffer_continuing_setter(int &argc, char *argv[], int w, int h,
                               int buffer_index = 0) :
    xcutbuffer_setter<B>(argc, argv, w, h, buffer_index) {}
  int loop()
  {
    char buf[64];       // small buffer for binary std::cin of pipe
    std::ios::sync_with_stdio(false);
    while (!0) {
      size_t size = xcutbuffer_setter<B>::data.size;
      struct timeval tv = { 0, 1000000/10 };
      struct timeval tv_start, tv_stop, tv_ZERO = { 0, 0 };
      gettimeofday(&tv_start, NULL);
      while (!std::cin.eof()) {
        std::streamsize size = sizeof(buf);
        std::cin.read(buf, size);
        size = std::cin.gcount();
        if (size)
          deject_if(!xcutbuffer_setter<B>::data.append(buf, size));
        gettimeofday(&tv_stop, NULL);
        if (tv < tv_stop - tv_start) { tv = tv_ZERO; break; }
      }
      if (xcutbuffer_setter<B>::data.size > size) {
        VERBOSE(std::cerr << xcutbuffer_setter<B>::data.size << " bytes read from stdin" << std::endl);
        xcutbuffer_setter<B>::store();
      }
      if (select(0, NULL, NULL, NULL, &tv) == -1) {}
      std::cin.clear();
    }
    return 0;
  }
};
template <typename B>
class xcutbuffer_following_setter : public xcutbuffer_continuing_setter<B> {
public:
  xcutbuffer_following_setter(int &argc, char *argv[],
                              int buffer_index = 0) :
    xcutbuffer_continuing_setter<B>(argc, argv, buffer_index) {}
  xcutbuffer_following_setter(int &argc, char *argv[], int w, int h,
                              int buffer_index = 0) :
    xcutbuffer_continuing_setter<B>(argc, argv, w, h, buffer_index) {}
  int loop()
  {
    std::cin.seekg(0, std::ios::end);
    std::cin.clear();
    return xcutbuffer_continuing_setter<B>::loop();
  }
};
template <typename B>
class xcutbuffer_continuing_adder : public xcutbuffer_continuing_setter<B> {
public:
  xcutbuffer_continuing_adder(int &argc, char *argv[],
                              int buffer_index = 0) :
    xcutbuffer_continuing_setter<B>(argc, argv, buffer_index) {}
  xcutbuffer_continuing_adder(int &argc, char *argv[], int w, int h,
                              int buffer_index = 0) :
    xcutbuffer_continuing_setter<B>(argc, argv, w, h, buffer_index) {}
  int loop()
  {
    xcutbuffer_getter<B>::fetch();
    xcutbuffer_continuing_setter<B>::loop();
    return 0;
  }
};
template <typename B>
class xcutbuffer_following_adder : public xcutbuffer_following_setter<B> {
public:
  xcutbuffer_following_adder(int &argc, char *argv[],
                             int buffer_index = 0) :
    xcutbuffer_following_setter<B>(argc, argv, buffer_index) {}
  xcutbuffer_following_adder(int &argc, char *argv[], int w, int h,
                             int buffer_index = 0) :
    xcutbuffer_following_setter<B>(argc, argv, w, h, buffer_index) {}
  int loop()
  {
    xcutbuffer_getter<B>::fetch();
    xcutbuffer_following_setter<B>::loop();
    return 0;
  }
};
template <typename B>
class xcutbuffer_rotator : public B {
public:
  int rotation_number;
  xcutbuffer_rotator(int &argc, char *argv[],
                     int rotation_number_ = 1) :
    B(argc, argv), rotation_number(rotation_number_) {}
  xcutbuffer_rotator(int &argc, char *argv[], int w, int h,
                     int rotation_number_ = 1) :
    B(argc, argv, w, h), rotation_number(rotation_number_) {}
  int loop()
  {
    XRotateBuffers(B::display, rotation_number);
    return 0;
  }
};
template <typename B>
class xcutbuffer_swapper {
public:
  int &argc;
  char **argv;
  xcutbuffer_swapper(int &argc_, char *argv_[]) :
    argc(argc_), argv(argv_) {}
  int loop()
  {
    xcutbuffer_setter<B> v[] = {
      xcutbuffer_setter<B>(argc, argv, 0),
      xcutbuffer_setter<B>(argc, argv, 1),
    };
    const size_t n = sizeof(v)/sizeof(v[0]);
    size_t i;
    for (i=0; i<n; ++i) {
      v[i].xcutbuffer_getter<B>::fetch();
    }
    const int rotation_number = 1;
    for (i=0; i<n; ++i) {
      v[i].buffer_index = (rotation_number + ((rotation_number < 0) ? (-rotation_number/n+1)*n : 0) + i) % n;
      v[i].store();
    }
    return 0;
  }
};
#endif
